"""parsers

Retrieve a README file from awesome repository and parse it to get awesome list.
"""
import os
import re
import sys
from abc import abstractmethod

import requests

from awesome.cache import exists, load_readme, save_readme

sys.path.append(os.path.abspath(os.path.pardir))


class AbstractAwesomeParser(object):
    """An abstract parser class for providing the common methods for parsing the markdown based README

    Each Awesome README has slightly different formats,
        So should implements each parsers corresponding to each languages or topics
    """
    AWESOME_TITLE = ''
    AWESOME_README_URL = ''

    def __init__(self, force_request=False):
        """Initialize the parser

        Args:
            force_request: If true, the parser does not use cached, but always request the README from remote repository
        """
        self._markdown_header_regex = re.compile('#+\s+(.*)')
        self._markdown_link_line_regex = re.compile('\s?\[([^\[\]<>]+)\]\s?\(([^\(\)]+)\)(?: [-—] (.*))?')
        self.force_request = force_request

    def is_cached(self):
        """Check if there exists cached README"""
        return exists(self.AWESOME_TITLE)

    def from_cache(self):
        """Retrieve the cached README from cache directory"""
        return load_readme(self.AWESOME_TITLE)

    def cache(self, readme):
        # Cache the readme to cache directory
        save_readme(self.AWESOME_TITLE, readme)

    def read_readme(self):
        """Read the README content from remote awesome repository"""
        if self.is_cached() and not self.force_request:
            print('Fetching the README from cache...')
            readme = self.from_cache()
            print('Done.')
            return readme
        print('Fetching the README from awesome place ...')
        response = requests.get(self.AWESOME_README_URL)
        print('Done.')
        if response.status_code == 200:
            readme = response.content.decode('utf8')
            self.cache(readme)
            return readme
        raise requests.RequestException('Error occurs when getting the README from {}'.format(self.AWESOME_README_URL))

    def parse(self):
        """Parse the all content of README of Awesome repository"""
        content = self.find_content()
        awesome_blocks = self.parse_awesome_content(content)
        return awesome_blocks

    @abstractmethod
    def find_content(self):
        """Find only awesome content from README content"""
        raise NotImplementedError

    @abstractmethod
    def parse_awesome_content(self, content):
        """Parses the raw awesome content and returns a structured list as following form:

        The README is formed as follow:

        ------------------------------------
        ... table of content ...              <- TOC

        {separator between TOC and content}   <- Separator such as '----'

        ## category                            <- Awesome content
            * [title](link) - description      <-
            * [title](link) - description      <-
            * [title](link) - description      <-

        # others                               <- Useless parts
            * ...                              <-
        ------------------------------------

        awesome_blocks = [
            {
                'type': <'category'|'awesome'>,
                'line': 'line text',
                'link'; [optional],
            },
        ]

        Args
            content: All content of the awesome README
        """
        raise NotImplementedError

    def parse_category_title(self, line):
        """Parse the header to get category title"""
        title = self._markdown_header_regex.findall(line)[0]
        plain_title = '[{}]'.format(title)
        return plain_title

    def parse_link_line(self, line):
        """Parse the line which has awesome link, title or description"""
        name, link, desc = self._markdown_link_line_regex.findall(line)[0]
        if desc:
            plain_line = '[{}] - {}'.format(name, desc)
        else:
            plain_line = '[{}]'.format(name)
        return plain_line, link



from .parsers.awesome import AwesomeParser
from .parsers.elixir import AwesomeElixirParser
from .parsers.go import AwesomeGoParser
from .parsers.java import AwesomeJavaParser
from .parsers.javascript import AwesomeJavaScriptParser
from .parsers.php import AwesomePHPParser
from .parsers.python import AwesomePythonParser
from .parsers.ruby import AwesomeRubyParser
from .parsers.rust import AwesomeRustParser
from .parsers.scala import AwesomeScalaParser
from .parsers.swift import AwesomeSwiftParser


def load_parsers():
    return {
        'awesome': AwesomeParser,
        'elixir': AwesomeElixirParser,
        'go': AwesomeGoParser,
        'java': AwesomeJavaParser,
        'javascript': AwesomeJavaScriptParser,
        'php': AwesomePHPParser,
        'python': AwesomePythonParser,
        'ruby': AwesomeRubyParser,
        'rust': AwesomeRustParser,
        'scala': AwesomeScalaParser,
        'swift': AwesomeSwiftParser,
    }



import os

CACHE_DIRECTORY = os.path.join(os.path.expanduser('~'), '.awesome_cache')


def exists(awesome_title):
    """

    Args:
        awesome_title: Awesome repository title

    Returns:
        True if exists, False otherwise
    """
    awesome_cache_directory = os.path.join(CACHE_DIRECTORY, awesome_title)
    awesome_cached_readme = os.path.join(awesome_cache_directory, 'README.md')
    return os.path.exists(awesome_cached_readme)


def save_readme(awesome_title, readme):
    """Save the README on local cache directory

    Args:
        awesome_title: Awesome repository title
        readme: Readme content

    Returns:
        None
    """
    awesome_cache_directory = os.path.join(CACHE_DIRECTORY, awesome_title)
    awesome_cached_readme = os.path.join(awesome_cache_directory, 'README.md')
    if not os.path.exists(CACHE_DIRECTORY):
        os.mkdir(CACHE_DIRECTORY)
    if not os.path.exists(awesome_cache_directory):
        os.mkdir(awesome_cache_directory)
    with open(awesome_cached_readme, 'w+', encoding='utf8') as readme_md:
        readme_md.write(readme)


def load_readme(awesome_title):
    """Loads the cached readme content

    Args:
        awesome_title: Awesome repository title

    Returns:
        A string of cached readme content, None if there is no cached readme
    """
    awesome_cache_directory = os.path.join(CACHE_DIRECTORY, awesome_title)
    awesome_cached_readme = os.path.join(awesome_cache_directory, 'README.md')
    if not os.path.exists(awesome_cache_directory):
        raise NotADirectoryError('{} does not exists'.format(awesome_cache_directory))
    if not os.path.exists(awesome_cached_readme):
        raise FileNotFoundError('{} does not exists'.format(awesome_cached_readme))
    with open(awesome_cached_readme, 'r', encoding='utf8') as readme_md:
        return readme_md.read()



import io
import os
import sys
from shutil import rmtree

from setuptools import setup, find_packages, Command

import awesome

# Package meta-data.
NAME = 'awesome-finder'
DESCRIPTION = 'A TUI based awesome finder'
URL = 'https://github.com/mingrammer/awesome-finder'
EMAIL = 'k239507@gmail.com'

# What packages are required for this module to be executed?
REQUIRED = [
    'requests'
]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()


class PublishCommand(Command):
    """Support setup.py publish."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


setup(
    name=NAME,
    version=awesome.__version__,
    description=DESCRIPTION,
    long_description=long_description,
    author=awesome.__author__,
    author_email=EMAIL,
    url=URL,
    keywords='awesome finder curses tui',
    packages=find_packages(),
    entry_points={
        'console_scripts': ['awesome=awesome.__main__:main'],
    },
    install_requires=REQUIRED,
    include_package_data=True,
    license=awesome.__license__,
    python_requires='>=3',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    # $ setup.py publish support.
    cmdclass={
        'publish': PublishCommand,
    },
)



import webbrowser

import curses
import curses.textpad

from textwrap import shorten


class SearchScreen(object):
    """A window screen for searching the awesome list without browser"""
    UP = -1
    DOWN = 1

    query = ''

    def __init__(self, awesome_title, awesome_blocks):
        """ Initialize the screen window

        Args
            awesome_title: Awesome topic title
            awesome_blocks: A list of formatted awesome content.
                It has a set of names and web links for crawled awesome content.

        Attributes
            window: A full curses screen window
            result_window: A window for showing the results
            search_window: A window for search bar

            y: Current y coordinates

            width: The width of `window`
            height: The height of `window`

            awesome_title: Awesome title
            awesome_blocks: It holds given `awesome_blocks` argument
            matched_blocks: A list of found awesome content

            max_lines: Maximum visible line count for `result_window`
            top: Available top line position for current page (used on scrolling)
            bottom: Available bottom line position for whole pages (as length of found lines)
            current: Current highlighted line number

            query: Query string for searching the awesome content

        Returns
            None
        """
        self.window = None
        self.result_window = None
        self.search_window = None

        self.y = 0

        self.width = 0
        self.height = 0

        self.awesome_title = awesome_title
        self.awesome_blocks = awesome_blocks
        self.matched_blocks = awesome_blocks

        self.init_curses()
        self.init_layout()

        self.max_lines = curses.LINES - 4
        self.top = 0
        self.bottom = self.max_lines
        self.current = 0

        self.run()

    def init_curses(self):
        """Setup the curses"""
        self.window = curses.initscr()
        self.window.keypad(True)

        curses.noecho()
        curses.cbreak()

        curses.start_color()
        curses.init_pair(1, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_CYAN)

        self.current = curses.color_pair(2)

    def init_layout(self):
        """Initialize the each windows with their size and shape"""
        self.height, self.width = self.window.getmaxyx()

        # Title section
        self.window.addstr(0, 0, '[awesome-{}] Find awesome things!'.format(self.awesome_title), curses.color_pair(1))
        self.window.hline(1, 0, curses.ACS_HLINE, self.width)

        # Search result section
        self.result_window = curses.newwin(self.height - 4, self.width, 2, 0)
        self.result_window.keypad(True)

        # Search bar section
        self.window.hline(self.height - 2, 0, curses.ACS_HLINE, self.width)
        self.window.addch(self.height - 1, 0, '>')
        self.search_window = curses.newwin(1, self.width - 1, self.height - 1, 2)
        self.search_window.keypad(True)

        self.window.refresh()

    def reset_top(self):
        """Reset the top

        It is used to move up the scroll page to top position for next query
        """
        self.top = 0

    def run(self):
        """Continue running the awesome finder until interrupted"""
        try:
            self.input_stream()
        except KeyboardInterrupt:
            pass
        finally:
            curses.endwin()

    def input_stream(self):
        """Waiting an input and run a proper method according to type of input"""
        while True:
            self.search(self.query)
            self.display()

            ch = self.search_window.getch()
            if curses.ascii.isprint(ch):
                self.write(ch)
                self.reset_top()
            elif ch in (curses.ascii.BS, curses.ascii.DEL, curses.KEY_BACKSPACE):
                self.delete()
                self.reset_top()
            elif ch == curses.KEY_UP:
                self.scroll(self.UP)
            elif ch == curses.KEY_DOWN:
                self.scroll(self.DOWN)
            elif ch in (curses.ascii.LF, curses.ascii.NL):
                self.open_link()
            elif ch == curses.ascii.ESC:
                break

    def write(self, ch):
        """Write a character and append it to the query"""
        self.search_window.addch(ch)
        self.query += chr(ch)

    def delete(self):
        """Delete an ending character"""
        self.y, _ = self.search_window.getyx()
        if len(self.query) > 0:
            self.search_window.move(self.y, len(self.query) - 1)
        self.search_window.delch()
        self.query = self.query[:-1]

    def scroll(self, direction):
        """Scrolling the result window when pressing up/down arrow keys

        Args:
            direction: Up or Down

        Returns:
            None
        """
        next_line = self.current + direction

        # Paging
        if direction == self.UP and (self.current == 0 and self.top != 0):
            self.top += self.UP
            return
        if direction == self.DOWN and (next_line == self.max_lines) and (self.top + self.max_lines != self.bottom):
            self.top += self.DOWN
            return
        # Scrolling on result window
        if direction == self.UP and (self.current != 0 or self.top != 0):
            self.current = next_line
            return
        if direction == self.DOWN and (self.top + self.current + 1 != self.bottom) and (self.current != self.max_lines):
            self.current = next_line
            return

    def open_link(self):
        """Open the link of highlighted awesome content"""
        index = self.top + self.current
        if self.matched_blocks[index]['type'] == 'awesome':
            webbrowser.open(self.matched_blocks[index]['link'], new=2)

    def search(self, query):
        """Search the awesome content with given query

        Args:
            query: Query string

        Returns:
            None
        """
        matched_blocks = []
        if query:
            query_tokens = query.lower().split()            
            for block in self.awesome_blocks:
                if all([token in block['line'].lower() for token in query_tokens]):
                    matched_blocks.append(block)
        else:
            matched_blocks = self.awesome_blocks

        # Set the bottom to length of matched blocks
        self.bottom = len(matched_blocks)

        # When reached bottom line, stop scrolling
        if self.current > self.bottom:
            self.current = self.bottom - 1
        self.matched_blocks = matched_blocks

    def display(self):
        """Display the found awesome content on result window"""
        self.result_window.erase()
        for idx, val in enumerate(self.matched_blocks[self.top:self.top + self.max_lines]):
            if val['type'] == 'category':
                if idx == self.current:
                    self.result_window.addstr(idx, 0, shorten(val['line'], self.width, placeholder='...'),
                                              curses.color_pair(2))
                else:
                    self.result_window.addstr(idx, 0, shorten(val['line'], self.width, placeholder='...'),
                                              curses.color_pair(1))
            elif val['type'] == 'awesome':
                if idx == self.current:
                    self.result_window.addstr(idx, 2, shorten(val['line'], self.width - 3, placeholder='...'),
                                              curses.color_pair(2))
                else:
                    self.result_window.addstr(idx, 2, shorten(val['line'], self.width - 3, placeholder='...'))
        self.result_window.refresh()


#!/usr/local/bin/python3

import argparse
import os

from .parser_loader import load_parsers
from .tui import SearchScreen

awesome_parsers = load_parsers()


def set_short_esc_delay():
    """Make the ESC key delay short"""
    os.environ.setdefault('ESCDELAY', '50')


def get_awesome_blocks(awesome_title, force):
    parser = awesome_parsers[awesome_title](force_request=force)
    awesome_blocks = parser.parse()
    return awesome_blocks


def parse_command():
    parser = argparse.ArgumentParser(description='awesome command')

    subparsers = parser.add_subparsers(dest='title', title='title',
                                       description='The title of awesome you want to find')
    subparsers.required = True

    # Register awesome commands
    for title in awesome_parsers.keys():
        if title == 'awesome':
            subparsers.add_parser(title, help='Search the {}'.format(title)) \
                .add_argument('--force', '-f', type=bool, nargs='?', const=True, default=False)
        else:
            subparsers.add_parser(title, help='Search the awesome-{}'.format(title)) \
                .add_argument('--force', '-f', type=bool, nargs='?', const=True, default=False)

    return parser.parse_args()


def main():
    set_short_esc_delay()

    args = parse_command()
    awesome_blocks = get_awesome_blocks(args.title, args.force)

    SearchScreen(args.title, awesome_blocks)


if __name__ == '__main__':
    main()



from . import AbstractAwesomeParser


class AwesomeRubyParser(AbstractAwesomeParser):
    AWESOME_TITLE = 'ruby'
    AWESOME_README_URL = 'https://raw.githubusercontent.com/markets/awesome-ruby/master/README.md'

    def find_content(self):
        readme = self.read_readme()
        content = readme.split('* [Resources](#resources)')[1]
        lines = []
        for line in content.split('\n'):
            lines.append(line)
        return lines

    def parse_awesome_content(self, content):
        awesome_blocks = []
        for line in content:
            # Parse the header title
            if line.startswith('#'):
                plain_title = self.parse_category_title(line)
                awesome_blocks.append({
                    'type': 'category',
                    'line': plain_title,
                })
            # Parse the list item
            elif line.strip().startswith('* [') or line.strip().startswith('- ['):
                plain_line, link = self.parse_link_line(line)
                awesome_blocks.append({
                    'type': 'awesome',
                    'line': plain_line,
                    'link': link,
                })
        return awesome_blocks
